/*
 * This service coordinates the local storage and remote API to provide a
 * consistent interface for controllers interacting with Deck/Answer data
 * 
 * Whenever the stored decks are downloaded, "Deck::Updated" is broadcast to
 * the app's $rootScope
 */

/* global magooshServices */

magooshServices.factory('Deck', ['$q', '$rootScope', 'persistencejs', 'FlashcardApi', function($q, $rootScope, persistencejs, FlashcardApi){
  
  var sync = function(){
    console.log('sync');
    var deferred = $q.defer();
    FlashcardApi.listDecks().then(function(data){
      console.log("Successfully received deck data");
      persistencejs.storeDecks(data)
        .then(function(){
          $rootScope.$broadcast('Deck::Updated'); // Notify controllers
        })
        .then(deferred.resolve);
    }, deferred.reject);
    return deferred.promise;
  };
  
  var submittingAnswers = false; // lock on answer syncing
  var submitNextAnswerFromBuffer = function(success, error){
    if (submittingAnswers) {
      // try again in 200ms
      console.log("Already syncing answers");
      setTimeout(function(){
        submitNextAnswerFromBuffer(success, error);
      }, 200);
    } else {
      submittingAnswers = true;
      persistencejs.loadAnswer()
        .then(function(answer){
          if (answer) {
            FlashcardApi.createAnswer(answer.card.deck.deck_id, answer.card.card_id, answer.was_known).then(function(){
              persistencejs.removeAnswer(answer).then(function(){
                console.log("NEXT ANSWER");
                submittingAnswers = false;
                submitNextAnswerFromBuffer(success, error);
              });
            }, function(message){
              console.log("ERROR");
              submittingAnswers = false;
              error(message);
            });  
          } else {
            console.log("DONE");
            submittingAnswers = false;
            success();
          }
        });
    }
  };
  
  /*
   * Sends all answers buffered in the local database to the API, then resolves
   * or rejects the promise
   * 
   * Returns a promise which is resolved if all answers are successfully
   * submitted, and rejected otherwise
   */
  var clearAnswerBuffer = function(){
    console.log("Submitting all remaining answer data");
    var deferred = $q.defer();
    
    submitNextAnswerFromBuffer(deferred.resolve, deferred.reject);
    
    return deferred.promise;
  };
  
  return {
    /*
     * List all decks, downloading from the server if they aren't stored
     * locally yet
     *
     * The returned promise is resolved with an array of deck objects or
     * rejected with an error message
     */
    all: function(){
      var deferred = $q.defer();
      persistencejs.listDecks().then(function(decks){
        if(decks.length > 0){
          console.log("Using local deck data");
          deferred.resolve(decks);
        } else {
          console.log("Syncing deck data");
          sync().then(function(){
            persistencejs.listDecks(deferred.resolve);
          }, deferred.reject);
        }
      });
      return deferred.promise;
    },
    
    /*
     * Fetch a deck from the database by an id
     * 
     * deck_id - the deck_id value of the deck to find
     * 
     * The returned promise is resolved with a deck object or rejected if no
     * deck was found
     */
    find: function(deck_id){
      return persistencejs.findDeck(deck_id);
    },
    
    /*
     * Choose a card to show the user
     * 
     * deck - the deck to select a card from
     * 
     * The returned promise is resolved with the selected card object
     */
    pickCard: function(deck){
      return persistencejs.pickCardFromDeck(deck);
    },
    
    /*
     * Enqueue an answer to be sent to the API and select the next card to show
     * 
     * card     - the flashcard the answer is being submitted for
     * wasKnown - a boolean value indicating whether "I knew this..." was
     *            selected
     * 
     * The returned promise is resolved with a card object
     */
    submitAnswerAndPickCard: function(card, wasKnown){
      var deferred = $q.defer();
      persistencejs.addAnswer(card, wasKnown)
        .then(function(){
          // Start clearing buffered answers after UI is done with callbacks
          setTimeout(clearAnswerBuffer().resolve, 1000);
        })
        .then(function(){
          persistencejs.pickCardFromDeck(card.deck, card.card_id)
            .then(deferred.resolve, deferred.reject);
        });
      return deferred.promise;
    },
    
    /*
     * Clear all locally stored deck data and re-download it from the API
     * 
     * The returned promise is resolved when the new data is saved or rejected
     * with an error message
     */
    sync: function(){
      var deferred = $q.defer();
      clearAnswerBuffer()
        .then(function(){
          sync().then(deferred.resolve, deferred.reject);
        }, deferred.reject);
      return deferred.promise;
    },
    
    /*
     * Submit a reset request to the API and clear all of the local deck
     * data so that it must be re-synced
     * 
     * The returned promise is resolved when the API request is completed and
     * the data is reset or rejected with an error message
     */
    reset: function(){
      var deferred = $q.defer();
      clearAnswerBuffer()
        .then(function(){
          FlashcardApi.reset().then(function(){
            persistencejs.deleteDecks()
              .then(deferred.resolve, deferred.reject);
          }, deferred.reject);
        }, deferred.reject);
      return deferred.promise;
    },
    
    /*
     * Submit a restore request to the API and clear all of the local deck
     * data so that it must be re-synced
     * 
     * The returned promise is resolved when the API request is completed and
     * the data is reset or rejected with an error message
     */
    restore: function(){
      var deferred = $q.defer();
      clearAnswerBuffer()
        .then(function(){
          FlashcardApi.restore().then(function(){
            persistencejs.deleteDecks()
              .then(deferred.resolve, deferred.reject);
          }, deferred.reject);
        }, deferred.reject);
      return deferred.promise;
    }
  };
}]);
