/*
 * This service wraps all calls to Magoosh's Flashcard REST API
 *
 * Tokens for identifying the user are stored in the persistencejs service's
 * Auth table and used for API requests where needed.
 */

/* global magooshServices, magooshHost, flashcardDeckSetId */

magooshServices.factory('FlashcardApi', ['$http', '$q', '$mixpanel', 'persistencejs', function($http, $q, $mixpanel, persistencejs){
  var flashcardApiEndpoint = 'https://' + magooshHost + '/api/v2/flashcards/sets/' + flashcardDeckSetId;
  
  /*
   * Wrap a function inside a promise, providing that function with the auth
   * token, and resolving the promise with the result
   */
  var withToken = function(func){
    var deferred = $q.defer();
    persistencejs.getAuthToken()
      .then(function(token){
        console.log("Making API call with token: " + token);
        func(token).then(deferred.resolve, deferred.reject);
      }, function(){
        deferred.reject(["Not logged in"]);
      });
    return deferred.promise;
  };
  
  return {
    
    /*
     * Send credentials to the API to login to get the user's token. If they
     * don't already have an account, then an account can be created.
     *
     * email         - the user's email address
     * password      - the user's password
     * createAccount - set to true to create an account if there's no existing
     *                 account for the email, or set to false to require an
     *                 existing account
     *
     * The returned promise is resolved with the user data (an object with
     * "email" and "token" attributes) or rejected with a list of error message
     * strings
     */
    login: function(email, password, createAccount){
      return withToken(function(token){
        var deferred = $q.defer();
        var post_data = {
          login: (email || ''),
          password: (password || ''),
          create: createAccount,
          api_auth_token: token
        };
        $http.post(flashcardApiEndpoint + '/login.json', post_data)
          .success(function(data){
            if (createAccount) {
              $mixpanel.track("Create Account", {'Email': data.email});
            } else {
              $mixpanel.track("Login", {'Email': data.email});
            }
            var user = {
              email: data.email,
              token: data.api_auth_token
            };
            $mixpanel.setUserId(user.token);
            $mixpanel.alias(user.id, user.token);
            deferred.resolve(user);
          })
          .error(function(data){
            if (data && data.error_messages.length > 0){
              deferred.reject(data.error_messages);
            } else {
              deferred.reject(["Couldn't reach the server"]);
            }
          });
        return deferred.promise;
      });
    },
    
    /*
     * List all decks with their associated cards and current progress data
     *
     * The returned promise is resolved with the an array of deck objects or
     * rejected if there's an API error
     */
    listDecks: function(){
      return withToken(function(token){
        var deferred = $q.defer();
        $http.get(flashcardApiEndpoint + '/decks.json?include=cards&api_auth_token=' + token)
          .success(deferred.resolve)
          .error(function(data){
            deferred.reject(["Couldn't reach the server"]);
          });
        return deferred.promise;
      });
    },
    
    /*
     * Create a new answer
     *
     * deckId   - a deck's id
     * cardId   - a card's id
     * wasKnown - a boolean value indicating whether "I knew this..." was
     *            selected
     *
     * The returned promise is resolved if the answer is created successfully
     * or rejected if it wasn't created
     */
    createAnswer: function(deckId, cardId, wasKnown){
      return withToken(function(token) {
        var deferred = $q.defer();
        var post_data = {
          api_auth_token: token,
          deck_id: deckId,
          card_id: cardId,
          result: (wasKnown ? 'known' : 'not_known')
        };
        console.log("Send answer for card " + post_data.card_id + " to API");
        $http.post(flashcardApiEndpoint + '/answers.json', post_data)
          .success(deferred.resolve)
          .error(function(data){
            deferred.reject(["Couldn't reach the server"]);
          });
        return deferred.promise;
      });
    },
    
    /*
     * Reset all flashcard usage data. The user's local progress should be
     * reset and synced if this request succeeds
     *
     * The returned promise is resolved if the request succeeds and rejected if
     * there's an error
     */
    reset: function(){
      return withToken(function(token){
        var deferred = $q.defer();
        $http.put(flashcardApiEndpoint + '/reset.json?api_auth_token=' + token)
          .success(deferred.resolve)
          .error(function(data){
            deferred.reject(["Couldn't reach the server"]);
          });
        return deferred.promise
      });
    },
    
    /*
     * Restore all flashcard usage data. The user's local progress should be
     * reset and synced if this request succeeds
     *
     * The returned promise is resolved if the request succeeds and rejected if
     * there's an error
     */
    restore: function(){
      return withToken(function(token){
        var deferred = $q.defer();
        // Old versions of Android (e.g. on a Droid X) don't like ".delete" so we use a syntactic workaround
        $http['delete'](flashcardApiEndpoint + '/reset.json?api_auth_token=' + token)
          .success(deferred.resolve)
          .error(function(data){
            deferred.reject(["Couldn't reach the server"]);
          });
        return deferred.promise;
      });
    }
  };
}]);
