/*
 * This service provides a consistent interface for storing and reading data
 * from a WebSQL database and handling caching of remote resources
 * 
 * All of this service's methods return a promise object.
 *
 * Note: Since persistence.js uses asynchronous callbacks for all of its calls,
 * the digests are explicitly triggered with $rootScope.$apply
 */

/* global magooshServices, persistence, ImgCache */

magooshServices.service('persistencejs', ['$rootScope', '$q', function($rootScope, $q){
  // score constants
  var masteredLeitnerScore = 3;
  var firstAnswerScore = masteredLeitnerScore + 2;
  
  persistence.store.websql.config(persistence, 'flashcards_v2', 'Magoosh flashcards database', 5 * 1024 * 1024);
  var Auth = persistence.define('auth', {
    email: 'TEXT',
    token: 'TEXT'
  });
  var Deck = persistence.define('deck', {
    // data from API
    deck_id: 'INT',
    slug: 'TEXT',
    name: 'TEXT',
    card_name_singular: 'TEXT',
    card_name_plural: 'TEXT',
    card_name_back: 'TEXT',
    // local caching and calculation
    cards_count: 'INT',
    new_cards_count: 'INT',
    mastered_cards_count: 'INT',
    reviewing_cards_count: 'INT',
    learning_cards_count: 'INT'
  });
  var Card = persistence.define('card', {
    // data from API
    card_id: 'INT',
    slug: 'TEXT',
    front: 'JSON',
    back: 'JSON',
    result_value_sum: 'INT',
    // local caching and calculation
    learning_state: 'TEXT',
    leitner: 'INT',
    sources_cached: 'BOOL'
  });
  var Answer = persistence.define('answer', {
    was_known: 'BOOL'
  });
  var SessionData = persistence.define('session_data', {
    key: 'TEXT',
    value: 'JSON'
  });
  
  Deck.hasMany('cards', Card, 'deck');
  Card.hasMany('answers', Answer, 'card');
  
  // Save the original table definitions
  persistence.schemaSync();
  
  var clearDeckData = function(){
    var deferred = $q.defer();
    console.log('Clearing deck data');
    Deck.all().destroyAll(function(){
      Card.all().destroyAll(function(){
        Answer.all().destroyAll(function(){
          deferred.resolve();
          $rootScope.$apply(); // Start digest to resolve promises
        });
      });
    });
    return deferred.promise;
  };
  
  var clearUserData = function(){
    var deferred = $q.defer();
    console.log('Clearing all user data');
    Auth.all().destroyAll(function(){
      clearDeckData().then(function(){
        deferred.resolve();
        if(!$rootScope.$$phase){ $rootScope.$apply(); } // Start digest
      }, deferred.reject);
    });
    return deferred.promise;
  };
  
  /* Internal: Helper to determine whether a card has any remote sources which need local caching
   * 
   * card - the Card to check
   * 
   * Returns true if there are any entries with a source attribute, false otherwise
   */
  var cardHasEntrySources = function(card){
    var i;
    for (i = 0; i < card.front.length; i++){
      if (card.front[i].source) {return true;}
    }
    for (i = 0; i < card.back.length; i++){
      if (card.back[i].source) {return true;}
    }
    return false;
  };
  
  /* Caches all of the entries for a single card, persists to the database and
   * makes a callback
   *
   * card             - the card to cache
   * success_callback - called when all entries have their sources cached
   * error_callback   - called if there's an error and caching doesn't complete
   */
  var cacheCardSources = function(card, success_callback, error_callback){
    
    /* Internal: Helper for cacheCardSources to asynchronously iterate over a list
     * of entries, caching any remote sources to the local filesystem with ImgCache
     * 
     * entries       - an Array of card entries (from Card.front or Card.back)
     * index         - index within entries to start at (should be called set to 0)
     * cache_success - called when all entries have their sources cached
     * cache_error   - called if the caching doesn't complete
     */
    var cacheEntries = function(entries, index, cache_success, cache_error){
      var entry = entries[index]; // current entry is removed from entries array
      if (typeof entry === 'undefined'){
        cache_success();
      } else {
        if (entry.source){
          // helper to replace the content with the cached version and continue
          var replaceWithCachedFile = function(url){ 
            entry.content = entry.content.replace(entry.source, url);
            cacheEntries(entries, index + 1, cache_success, cache_error); // cache next entry
          };
          
          // Finds the path of the image in the cache. If it's not already cached,
          // then download the image and try again.
          ImgCache.isCached(function(path, success){
            if (success) {
              replaceWithCachedFile(path);
            } else {
              // Not already cached, so download the file and try again
              ImgCache.cacheFile(entry.source, function(){
                // Get the path again
                ImgCache.isCached(function(path, success){
                  if (success) {
                    replaceWithCachedFile(path);
                  } else {
                    cache_error();
                  }
                });
              }, cache_error);
            }
          });
        } else {
          cacheEntries(entries, index + 1, cache_success, cache_error); // cache next entry
        }
      }
    };
    
    console.log('caching card "' + card.slug + '"...');
    cacheEntries(card.front, 0, function(){
      card.markDirty('front');
      cacheEntries(card.back, 0, function(){
        card.markDirty('back');
        card.sources_cached = true;
        persistence.flush(success_callback); // done, persist and call success callback
      }, error_callback);
    }, error_callback);
  };
  
  // Iterates over the Card table and checks 
  // when a card has external sources, they are downloaded to the filesystem and 
  var cacheAllCardSources = function(card_scope){
    card_scope.filter('sources_cached', '==', false).one(function(card){
      if (card){
        cacheCardSources(card, function(){
          // recursively check the scope remaining uncached cards
          cacheAllCardSources(card_scope);
        }, function(){
          console.log('Could not download sources for card ' + card.card_id);
        });
      } else {
        return; // stop the recursive calls, no cards left
      }
    });
  };
  
  var updateLearningStateCaches = function(card){
    // Store the current learning_state for detecting changes for caches
    var previousLearningState = card.learning_state;
    
    // Cache the learning state
    if (card.result_value_sum === null){
      card.learning_state = 'new';
    } else if (card.result_value_sum >= masteredLeitnerScore) {
      card.learning_state = 'mastered';
    } else if (card.result_value_sum < 0) {
      card.learning_state = 'learning';
    } else {
      card.learning_state = 'reviewing';
    }
    
    if (previousLearningState !== card.learning_state){
      if(previousLearningState){
        card.deck[previousLearningState + '_cards_count']--;
        card.deck.markDirty(previousLearningState + '_cards_count');
      }
      card.deck[card.learning_state + '_cards_count']++;
      card.deck.markDirty(card.learning_state + '_cards_count');
    }
  };
  
  var sampleByLeitner = function(cards){
    var bestScore, curScore;
    var bestCard = cards[0];
    for(var i=0; i<cards.length; i++){
      // Lowest allowed value is -1 (ignore anything lower), distribution divides by ln(3) to get log3(rand)
      curScore = Math.max(cards[i].result_value_sum, -1) + Math.log(1 - Math.random()) / 1.0986;
      
      // if the card has unsynced remote sources, it's given a penalty weight so that
      // we probably won't show it to students
      if (!cards[i].sources_cached) {
        curScore += masteredLeitnerScore;
      }
      
      // We're finding the card with the lowest score
      if (!bestScore || curScore < bestScore){
        bestScore = curScore;
        bestCard = cards[i];
      }
    }
    return bestCard;
  };
  
  // Generate a random 40-character string
  var generateToken = function(){
    var possible = '0123456789abcdef';
    var text = '';
    for(var i = 0; i < 40; i++){
      text += possible.charAt(Math.floor(Math.random() * possible.length));
    }
    return text;
  };
  
  // singleton containing all methods to be called
  return {
    /*
     * Clear all user data and store a new user in the Auth table. If no user
     * object is provided, then the table will be left empty
     *
     * user - an object with "email" and "token" fields
     */
    setUser: function(user){
      var deferred = $q.defer();
      clearUserData().then(function(){
        // add user to the table, if provided
        if (user) {
          persistence.add(new Auth(user));
        }
        persistence.flush(function(){
          deferred.resolve();
          $rootScope.$apply(); // start digest
        });
      });
      return deferred.promise;
    },
    
    /*
     * Read the user data stored in the Auth table
     * 
     * The returned promise is resolved with the user data (a Javascript object
     * with "email" and "token" attributes) if the user is logged in, and
     * rejected otherwise
     */
    getUser: function(){
      var deferred = $q.defer();
      Auth.all().one(function(user){
        if (user && user.email){
          deferred.resolve(user);
        } else {
          deferred.reject();
        }
        $rootScope.$apply(); // start digest
      });
      return deferred.promise;
    },
    
    /*
     * Read the user's token stored in the Auth table or generate a new one if
     * they aren't logged in
     * 
     * The returned promise is resolved with the token
     */
    getAuthToken: function(){
      var deferred = $q.defer();
      Auth.all().one(function(auth){
        if (auth){
          // already persisted
          deferred.resolve(auth.token);
          $rootScope.$apply(); // start digest
        } else {
          // create a new token
          var anonymousAuth = new Auth({token: generateToken()});
          persistence.add(anonymousAuth);
          persistence.flush(function(){
            deferred.resolve(anonymousAuth.token);
            $rootScope.$apply();
          });
        }
      });
      return deferred.promise;
    },
    
    /*
     * Fetch a list of all of the decks in the Deck table
     * 
     * The returned promise is resolved with an array of deck objects
     */
    listDecks: function(){
      var deferred = $q.defer();
      Deck.all().list(function(decks){
        deferred.resolve(decks);
        $rootScope.$apply(); // start digest
      });
      return deferred.promise;
    },
    
    /*
     * Store a new set of decks in the Deck and Card tables. Each card and deck
     * is stored with cached progress values (cards_count, mastered_cards_count,
     * result_value_sum, etc.)
     * 
     * The returned promise is resolved when all of the data has been stored
     */
    storeDecks: function(decks){
      var deferred = $q.defer();
      
      clearDeckData().then(function(){
        console.log("Storing " + decks.length + " decks...");
        for(var i=0; i < decks.length; i++){
          var deck = new Deck({
            deck_id: decks[i].id,
            slug: decks[i].slug,
            name: decks[i].name,
            card_name_singular: decks[i].card_name_singular,
            card_name_plural: decks[i].card_name_plural,
            card_name_back: decks[i].card_name_back,
            cards_count: decks[i].cards.length,
            new_cards_count: 0,
            mastered_cards_count: 0,
            reviewing_cards_count: 0,
            learning_cards_count: 0
          });
          persistence.add(deck);
          for(var j=0; j < deck.cards_count; j++){
            var c = decks[i].cards[j];
            var card = new Card({
              card_id: c.id,
              slug:    c.slug,
              front:   c.front,
              back:    c.back,
              result_value_sum: c.result_value_sum
            });
            card.deck = deck;
            card.sources_cached = !cardHasEntrySources(card);
            updateLearningStateCaches(card);
            persistence.add(card);
          }
          console.log(deck.name + ' has ' + deck.cards_count + ' cards');
        }
        persistence.flush(function(){
          deferred.resolve();
          $rootScope.$apply(); // Start digest to resolve promises
          
          // Start remote source caching
          cacheAllCardSources(Card.all());
        });
      });
      return deferred.promise;
    },
    
    /*
     * Delete all stored deck data
     */
    deleteDecks: function(){
      return clearDeckData();
    },
    
    /*
     * Fetch a deck from the database by an id
     * 
     * deck_id - the deck_id value of the deck to find
     * 
     * The returned promise is resolved with a deck object or rejected if no
     * deck was found
     */
    findDeck: function(deck_id){
      var deferred = $q.defer();
      Deck.findBy('deck_id', deck_id, function(deck){
        if (deck){
          deferred.resolve(deck);
        } else {
          deferred.reject();
        }
        $rootScope.$apply(); // start digest
      });
      return deferred.promise;
    },
    
    /*
     * Pick the next card from a deck for the user
     * 
     * deck            - the deck to select a card from
     * exclude_card_id - an optional card_id of a card record to be excluded
     *                   from the selection
     * 
     * The returned promise is resolved with the chosen card
     */
    pickCardFromDeck: function(deck, exclude_card_id){
      var deferred = $q.defer();
      
      var query = deck.cards;
      if (exclude_card_id) {
        query = query.filter('card_id', '!=', exclude_card_id);
      }
      query.list(function(cards){
        // Trigger the remote source sync in case the user interrupted the
        // original download
        cacheAllCardSources(deck.cards);
        
        deferred.resolve(sampleByLeitner(cards));
        $rootScope.$apply(); // start digest
      });
      return deferred.promise;
    },
    
    /*
     * Add a new answer to the queue in the Answer table and update the
     * progress caches on the card and deck records
     * 
     * card     - the flashcard the answer is being submitted for
     * wasKnown - a boolean value indicating whether "I knew this..." was
     *            selected
     *
     * The returned promise is resolved after the answer has been stored
     */
    addAnswer: function(card, wasKnown){
      var deferred = $q.defer();
      
      // New answer record
      var answer = new Answer({was_known: wasKnown});
      persistence.add(answer);
      card.answers.add(answer);
      
      // Calculate new result_value_sum for the card
      if (wasKnown) {
        if (card.result_value_sum === null){
          card.result_value_sum = firstAnswerScore; // first time, send directly to "mastered"
        } else {
          card.result_value_sum += 1; // previous answers, increment score
        }
      } else {
        card.result_value_sum = -1; // incorrect, send to "learning"
      }
      
      // Update count caches on the deck
      updateLearningStateCaches(card);
      
      persistence.flush(function(){
        deferred.resolve();
        $rootScope.$apply();
      });
      
      return deferred.promise;
    },
    
    /*
     * Load a single answer from the Answer table
     * 
     * The returned promise is resolved with the found answer (or null if
     * there were none)
     */
    loadAnswer: function(){
      var deferred = $q.defer();
      Answer.all().prefetch('card').one(function(answer){
        if (answer){
          // If there's an answer, we need to prefetch its deck too
          answer.card.fetch('deck', function(){
            deferred.resolve(answer);
            $rootScope.$apply(); // Start digest to resolve promises
          });
        } else {
          deferred.resolve(answer);
          $rootScope.$apply(); // Start digest to resolve promises
        }
      });
      return deferred.promise;
    },
    
    /*
     * Delete an answer from the Answer table
     * 
     * answer - the answer to delete
     * 
     * The returned promise is resolved after the record has been deleted
     */
    removeAnswer: function(answer){
      var deferred = $q.defer();
      persistence.remove(answer);
      persistence.flush(function(){
        deferred.resolve();
        $rootScope.$apply();
      });
      return deferred.promise;
    },


    /*
     * Get the value associated with a key
     *
     * key - string name of the key
     *
     * The returned promise is resolved with the value or returns
     # null if no matching key/value pair is found
     */
    getSessionData: function(key) {
      var deferred = $q.defer();
      SessionData.findBy('key', key, function(pair) {
        if (pair) {
          deferred.resolve(pair.value);
        } else {
          deferred.resolve(null)
        }
        $rootScope.$apply(); // Start digest
      });
      return deferred.promise;
    },

    /*
     * Update or set a key, value pair in the extras column object
     *
     * key   - string name of the key
     * value - the value to store, can be of any Javascript type
     *
     * The returned promise is resolved when the SessionData entry
     * is updated or created
     */
    setSessionData: function(key, value) {
      var deferred = $q.defer();
      SessionData.findBy('key', key, function(pair) {
        if (pair) {
          pair.value = value;
        } else {
          data = {
            key: key,
            value: value
          };
          persistence.add(new SessionData(data));
        }
        persistence.flush(function() {
          deferred.resolve();
          $rootScope.$apply(); // Start digest
        });
      });
      return deferred.promise;
    }
  };
}]);
